/*
 * bs.c
 * Copyright 2004 Deekoo L.
 *
 * It's a backspace proxy to let stupid client programs that only do raw
 * telnet communicated properly with stupid servers that only do linemode
 * telnet.
 *
 * It's probably got bugs.
 *
 * It's released to the general public under version 2.0 of the GNU GPL.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#ifdef WIN32
#  include <winsock2.h>
#else
#  include <sys/resource.h>
#  include <sys/select.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#endif

#define BUFMAX 10480

struct connection
{
        char inbuf[BUFMAX];
        int inpos;
        int inthrough;
        int insendend;
        int infd;
        struct sockaddr inaddr;
        
        char outbuf[BUFMAX];
        int outpos;
        int outthrough;
        int outfd;
};
struct sockaddr_in g_destaddr;

int g_numconns=0;
#define G_MAXCONNS 10
struct connection *g_conns[G_MAXCONNS];
fd_set g_wfds,g_rfds;
int g_highfd;

void die_help()
{
        fprintf(stderr,"Syntax:\nbs listen-ip listen-port dest-ip dest-port\nUse 0.0.0.0 as listen-ip to listen on all IP addresses to which your machine\nis currently connected.\n");
        exit(1);
}

#define TOOMANYCONNS "Too many connections.\nClosing.\n"

struct connection *accept_conn(int sock)
{
        struct connection *conn;
        conn = malloc(sizeof(struct connection));
        int x = sizeof(struct sockaddr_in);
        if (!conn) {
                fprintf(stderr,"malloc() failed.\n");
                exit(1);
        }
        conn->infd = accept(sock,NULL,&x);
        if (conn->infd<0) {
                fprintf(stderr,"Error %d (%s) accept()ing.\n",errno,strerror(errno));
                free(conn);
                return NULL;
        }
        if (g_numconns>G_MAXCONNS) {
                send(conn->infd,TOOMANYCONNS,strlen(TOOMANYCONNS),0);
                goto closeout;
        }
        g_conns[g_numconns]=conn;
        g_numconns++;
        conn->outfd = socket(PF_INET,SOCK_STREAM,0);
        if (conn->outfd<0) {
                fprintf(stderr,"Error %d connecting to destination.\n",errno);
                // XXX write error
                goto closeout;
        }
        if (connect(conn->outfd,(struct sockaddr *)&g_destaddr,sizeof(struct sockaddr_in))<0) {
                goto closeboth;
        }
#ifdef WIN32
        x=1;
        if (ioctlsocket(conn->infd,FIONBIO,&x)<0) {
                fprintf(stderr,"fionbio failed.\n");
        }
        if (ioctlsocket(conn->outfd,FIONBIO,&x)<0) {
                fprintf(stderr,"fionbio failed.\n");
        }
#else
        fcntl(conn->infd,F_SETFL,O_NONBLOCK);
        fcntl(conn->outfd,F_SETFL,O_NONBLOCK);
#endif
        
        conn->inpos = conn->inthrough = conn->insendend =
                conn->outpos = conn->outthrough = 0;
        FD_SET(conn->infd,&g_rfds);
        FD_SET(conn->outfd,&g_rfds);
        if (conn->infd>=g_highfd) g_highfd=conn->infd+1;
        if (conn->outfd>=g_highfd) g_highfd=conn->outfd+1;
        return conn;
closeboth:
        close(conn->outfd);
closeout:
        close(conn->infd);
        free(conn);
        return NULL;
}

void mark_writable(int fd)
{
        if (fd>=g_highfd) g_highfd=fd+1;
        FD_SET(fd,&g_wfds);
}

void kill_connection(struct connection *conn, const char *reason)
{
        int x=0;
        send(conn->infd,reason,strlen(reason),0);
        FD_CLR(conn->infd,&g_rfds);
        FD_CLR(conn->infd,&g_wfds);
        FD_CLR(conn->outfd,&g_rfds);
        FD_CLR(conn->outfd,&g_wfds);
        close(conn->infd);
        close(conn->outfd);
        while (x<g_numconns) {
                if (g_conns[x]==conn) {
                        g_conns[x]=g_conns[--g_numconns];
                }
                x++;
        }
        free(conn);
}

void write_str(int fd, const char *s)
{
        send(fd,s,strlen(s),0);
}

int conn_read_input(struct connection *conn)
{
        int x = conn->inpos;
        int numread;
        int numtoread=BUFMAX-x;
        int y,z,bs;
        int last;
#ifdef WIN32
        ioctlsocket(conn->infd,FIONREAD,&numtoread);
#endif
        numread = recv(conn->infd,&conn->inbuf[x],numtoread,0);
        if (numread<=0) {
                if (numread==0) {
                        if (x==BUFMAX) {
                                if (conn->inthrough) {
                                        y = 0;
                                        z = conn->inthrough;
                                        while (z<BUFMAX) {
                                                conn->inbuf[y++] = conn->inbuf[z++];
                                        }
                                        conn->insendend -= conn->inthrough;
                                        conn->inpos -= conn->inthrough;
                                        conn->inthrough = 0;
                                } else {
                                        if (!conn->insendend) {
                                                write_str(conn->infd,"Line too long.  Trimmed last character off.\n");
                                                conn->inpos--;
                                                conn->inthrough--;
                                        } else {
                                                fprintf(stderr,"Input buffer full.\n");
                                        }
                                }
                        } else {
                                kill_connection(conn,"Read 0 bytes.\n");
                        }
                } else {
                        kill_connection(conn,"Error reading into input buffer.\n");
                        return 0;
                }
        }
        conn->inpos+=numread;
        while (x<conn->inpos) {
                if (conn->inbuf[x]=='\010' || conn->inbuf[x]=='\177') {
                        y=z=x;
                        z++;
                        if (x>0 && conn->inbuf[y-1]!='\012' && 
                                        conn->inbuf[y-1]!='\015' &&
                                        conn->inthrough<y) {
                                y--;
                                bs = 1;
                        } else {
                                bs = 0;
                        }
                        while (z<conn->inpos) {
                                conn->inbuf[y++]=conn->inbuf[z++];
                        }
                        conn->inpos-=1+bs;
                        x-=1+bs;
               } else if (conn->inbuf[x]=='\012' || conn->inbuf[x]=='\015') {
                       last = conn->inbuf[x];
                       conn->inbuf[x]='\0';
                       mark_writable(conn->outfd);
                       conn->insendend = x;
                       conn->inbuf[x]=last;
//               } else if (conn->inbuf[x]=='\025') {
                       // XXX ^U
//               } else {
               }
               x++;
       }
       return numread;
}

void conn_write_input(struct connection *conn)
{
        int wrote;
        if (conn->inthrough<conn->insendend) {
                wrote = send(conn->outfd,&conn->inbuf[conn->inthrough],conn->insendend-conn->inthrough,0);
                if (wrote>0) {
                        conn->inthrough+=wrote;
                } else if (wrote<0) {
                        fprintf(stderr,"Error writing to conn->outfd.\n");
                }
        }
        if (conn->inthrough>=conn->insendend) {
                FD_CLR(conn->outfd,&g_wfds);
        }
}

void conn_write_output(struct connection *conn)
{
        int wrote;
        if (conn->outthrough<conn->outpos) {
                wrote = send(conn->infd,&conn->outbuf[conn->outthrough],
                                conn->outpos-conn->outthrough,0);
                if (wrote>0) {
                        conn->outthrough+=wrote;
                } else if (wrote<0) {
                        fprintf(stderr,"Error writing to conn->infd\n");
                        kill_connection(conn,"Error writing to input fd.\n");
                }
        }
        if (conn->outthrough>=conn->outpos) {
                FD_CLR(conn->infd,&g_wfds);
        }
}

void conn_read_output(struct connection *conn)
{
        int x,y;
        int numread;
        int numtoread = BUFMAX-conn->outpos;
#ifdef WIN32
        numtoread = 1;
#endif
        numread = recv(conn->outfd,&conn->outbuf[conn->outpos],BUFMAX-conn->outpos,0);
        if (numread>0) {
                conn->outpos += numread;
                FD_SET(conn->infd,&g_wfds);
        } else {
                if (numread<0) {
#ifdef WIN32
                        fprintf(stderr,"Errno: %d\n",WSAGetLastError());
#endif
                        kill_connection(conn,"Error reading from output server.");
                } else {
                        if (conn->outpos==BUFMAX) {
                                if (conn->outthrough) {
                                        x = 0;
                                        y = conn->outthrough;
                                        while (y<BUFMAX) {
                                                conn->outbuf[x++]=conn->outbuf[y++];
                                        }
                                        conn->outpos -= conn->outthrough;
                                        conn->outthrough = 0;
                                } else {
                                        FD_CLR(conn->outfd,&g_rfds);
                                        fprintf(stderr,"Output buffer full.\n");
                                }
                        } else {
                                kill_connection(conn,"Remote closed connection.\n");
                        }
                }
        }
}

int main(int argc, char **argv)
{
        fd_set readable,writable;
        int sock, selected, x;
        struct connection *conn;
        struct sockaddr_in listenaddr;
#ifdef WIN32
        WSADATA data;
        WSAStartup(MAKEWORD(1,1),&data);
#endif
        sock = socket(PF_INET, SOCK_STREAM, 0);

        if (argc != 5) {
                fprintf(stderr,"%d args.\n",argc);
                die_help();
        }
        if (sock<0) {
#ifdef WIN32
                errno=WSAGetLastError();
#endif
                fprintf(stderr,"Socket error (%d/%s).\n",errno,strerror(errno));
                exit(1);
        }
        listenaddr.sin_family = g_destaddr.sin_family = AF_INET;
        listenaddr.sin_addr.s_addr = inet_addr(argv[1]);
        listenaddr.sin_port = htons(atoi(argv[2]));
        g_destaddr.sin_addr.s_addr = inet_addr(argv[3]);
        g_destaddr.sin_port = htons(atoi(argv[4]));
        if ((x=bind(sock,(struct sockaddr *)&listenaddr,sizeof(struct sockaddr_in)))<0) {
#ifdef WIN32
                errno=WSAGetLastError();
#endif
                fprintf(stderr,"bind()=%d error %d\n",x,errno);
                exit(1);
        }
        if (listen(sock,5)<0) {
                fprintf(stderr,"Error %d(%s) listening.\n",errno,strerror(errno));
                exit(1);
        }
        FD_ZERO(&g_rfds);
        g_wfds = g_rfds;
        FD_SET(sock,&g_rfds);
        g_highfd = sock+1;
        while (1) {
                readable = g_rfds;
                writable = g_wfds;
                selected = select(g_highfd,&readable,&writable,NULL,NULL);
                // XXX add a timeout.
                if (selected<0) {
                        fprintf(stderr,"select error %d.\n",errno);
                        exit(1);
                }
                if (FD_ISSET(sock,&readable)) {
                        conn = accept_conn(sock);
                        fprintf(stderr,"accepted.\n");
                }
                x = 0;
                while (x<g_numconns) {
                        if (FD_ISSET(g_conns[x]->infd,&readable)) {
                                conn_read_input(g_conns[x]);
                        }
                        if (FD_ISSET(g_conns[x]->outfd,&writable)) {
                                conn_write_input(g_conns[x]);
                        }
                        if (FD_ISSET(g_conns[x]->outfd,&readable)) {
                                conn_read_output(g_conns[x]);
                        }
                        if (FD_ISSET(g_conns[x]->infd,&writable)) {
                                conn_write_output(g_conns[x]);
                        }
                        x++;                
                }
        }
}
